part of 'api.dart';

@immutable
class ApiResponse<T> {
  final bool success;
  final String? error;
  final T? value;
  final Exception? exception;
  final Response? source;

  const ApiResponse({
    this.success = true,
    this.value,
    this.source,
    this.error,
    this.exception,
  });

  bool get isUnauthorized =>
      !success &&
      ((exception as DioException?)?.response?.isUnauthorized ?? false);

  //TODO: only cater gql now, pending for REST
  String? get errorCode =>
      success ? null : ((exception as DioException?)?.response?.gqlErrorCode);

  static ApiResponse<T> fromResponse<T>(
    Response<dynamic> response, {
    T? data,
    String? error,
  }) {
    return ApiResponse(
      value: data ?? response.data,
      source: response,
      success: error == null,
      error: error,
    );
  }

  static ApiResponse<T> fromException<T>(Exception exception) {
    return ApiResponse(
      success: false,
      error: _parse(exception),
      exception: exception,
    );
  }

  static String _parse(Exception exception) {
    final dioError = exception as DioException?;
    if (dioError != null && dioError.response != null) {
      final response = dioError.response!;
      final data = response.data;
      String? message;
      if (response.gqlErrorCode != null) {
        message = response.gqlErrorMessage;
      } else if (data is Map) {
        message = data['message'];
      }
      if (message.isNotNullOrEmpty) return message!;
      return '${response.statusCode}: ${response.statusMessage}';
    }
    return exception.toString();
  }

  @override
  String toString() {
    return error ?? value?.toString() ?? 'ApiResponse(value: null)';
  }
}

extension FutureApiResponseHelper<T> on Future<ApiResponse<T>> {
  Future<ApiResponse<T>> onSuccess(void Function(T value) onValue,
      {void Function(Exception error)? orElse}) async {
    var response = await this;
    if (response.success) {
      onValue(response.value as T);
    } else {
      orElse?.call(response.exception!);
    }
    return response;
  }
}
