import 'package:viqcore_flutter/src/http_client/utils/dio_helper.dart';

import '../app.dart';

/// A callback that returns a Dio response, presumably from a Dio method
/// it has called which performs an HTTP request, such as `dio.get()`,
/// `dio.post()`, etc.
typedef HttpLibraryMethod<T> = Future<Response<T>> Function();

typedef ApiResponseTransformer<T> = T Function(dynamic data);

class ApiService {
  ApiService({required this.dio});

  final Dio dio;

  /// HTTP GET request.
  Future<ApiResponse<T>> get<T>(
    String path, {
    Map<String, dynamic>? queryParameters,
    Options? options,
    CancelToken? cancelToken,
    ProgressCallback? onReceiveProgress,
  }) {
    return request(
      'GET',
      path,
      queryParameters: queryParameters,
      options: options,
      cancelToken: cancelToken,
      onReceiveProgress: onReceiveProgress,
    );
  }

  /// HTTP POST request.
  Future<ApiResponse<T>> post<T>(
    String path, {
    dynamic data,
    Map<String, dynamic>? queryParameters,
    Options? options,
    CancelToken? cancelToken,
    ProgressCallback? onSendProgress,
    ProgressCallback? onReceiveProgress,
  }) {
    return request(
      'POST',
      path,
      data: data,
      queryParameters: queryParameters,
      options: options,
      cancelToken: cancelToken,
      onSendProgress: onSendProgress,
      onReceiveProgress: onReceiveProgress,
    );
  }

  Future<ApiResponse<T>> request<T>(String method, String path,
      {dynamic data,
      Map<String, dynamic>? queryParameters,
      Options? options,
      CancelToken? cancelToken,
      ProgressCallback? onSendProgress,
      ProgressCallback? onReceiveProgress,
      ApiResponseTransformer? transformer}) {
    return _mapApiResponse<T, dynamic>(
        () => dio.request(
              path,
              data: data,
              queryParameters: queryParameters,
              options: (options ?? Options())..method = method,
              cancelToken: cancelToken,
              onSendProgress: onSendProgress,
              onReceiveProgress: onReceiveProgress,
            ),
        transformer: transformer);
  }

  Future<ApiResponse<T>> gql<T>(String method, dynamic query,
      {String? path,
      Map<String, dynamic>? variables,
      Options? options,
      CancelToken? cancelToken,
      ProgressCallback? onSendProgress,
      ProgressCallback? onReceiveProgress,
      ApiResponseTransformer? transformer}) {
    String bodyQuery;

    if (query is GraphQLQuery) {
      bodyQuery = query.build();
    } else {
      bodyQuery = '$query';
    }

    final body = '$method { $bodyQuery }';

    options ??= Options();
    options.extra ??= {};
    options.extra!.addAll({'GQL': true});

    return _mapApiResponse<T, dynamic>(
        () => dio.post(
              path ?? '/graphql',
              data: {'query': body, 'variables': variables},
              options: options,
              cancelToken: cancelToken,
              onSendProgress: onSendProgress,
              onReceiveProgress: onReceiveProgress,
            ),
        transformer: transformer);
  }

  Future<ApiResponse<T>> _mapApiResponse<T, R>(HttpLibraryMethod<R> method,
      {ApiResponseTransformer? transformer}) async {
    try {
      var response = await method();
      var data = response.data;
      return ApiResponse.fromResponse(
        response,
        error: response.gqlErrors ? response.gqlErrorMessage : null,
        data: transformer?.call(data) ?? data,
      );
    } on Exception catch (exception) {
      return ApiResponse.fromException(exception);
    }
  }

  // ignore_for_file: always_put_control_body_on_new_line

  /// Join parts of a url path into a url
  ///
  /// Typically part1 will be a http://xxx type prefix
  /// and part2-part(n) are paths to be appended to the part1
  /// When concatenating parts the '/' separator will be stripped or
  /// added as necessary to unsure there is just one '/' between parts.
  String getUrl(String part,
      [String? part2,
      String? part3,
      String? part4,
      String? part5,
      String? part6,
      String? part7,
      String? part8,
      String? part9]) {
    var endpoint = _urlStrip(dio.options.baseUrl);

    endpoint = _urlJoin2(endpoint, part);
    if (part2 != null) endpoint = _urlJoin2(endpoint, part2);
    if (part3 != null) endpoint = _urlJoin2(endpoint, part3);
    if (part4 != null) endpoint = _urlJoin2(endpoint, part4);
    if (part5 != null) endpoint = _urlJoin2(endpoint, part5);
    if (part6 != null) endpoint = _urlJoin2(endpoint, part6);
    if (part7 != null) endpoint = _urlJoin2(endpoint, part7);
    if (part8 != null) endpoint = _urlJoin2(endpoint, part8);
    if (part9 != null) endpoint = _urlJoin2(endpoint, part9);
    return endpoint;
  }

  String _urlJoin2(String part1, String part2) =>
      '${_urlStrip(part1)}/${_urlStrip(part2)}';

  String _urlStrip(String part1) {
    String endpoint;
    if (part1.startsWith('/')) {
      endpoint = part1.substring(1);
    } else {
      endpoint = part1;
    }
    if (endpoint.endsWith('/')) {
      endpoint = endpoint.substring(0, endpoint.length - 1);
    }
    return endpoint;
  }
}
