// ignore_for_file: avoid_returning_this

part of 'api.dart';

class GraphQLQuery {
  final String name;
  final List<dynamic> fields;
  final Map<String, dynamic>? params;
  ApiService? _client;

  Map<String, String>? _headers;
  Map<String, dynamic>? _extra;
  Duration? _timeout;
  // final Map<String, dynamic>? _params;
  String? _method;

  final List<GraphQLQuery> _gqls = [];

  GraphQLQuery(this.name, this.fields, {this.params});

  GraphQLQuery external({ApiService? client}) {
    _client = client ?? App.getIt<API>();
    return this;
  }

  GraphQLQuery timeout(Duration value) {
    _timeout = value;
    return this;
  }

  GraphQLQuery headers(Map<String, String> value) {
    _headers = value;
    return this;
  }

  GraphQLQuery extra(Map<String, String> value) {
    _extra = value;
    return this;
  }

  GraphQLQuery combine(GraphQLQuery gql) {
    _gqls.add(gql);
    return this;
  }

  String _parseFields(List fieldList) {
    return fieldList.isEmpty
        ? _method == 'query'
            ? ''
            : 'void'
        : fieldList
            .map((x) {
              if (x is String) {
                return x;
              }
              if (x is GraphQLQuery) {
                return x.build();
              }
              if (x is List) {
                return _parseFields(x);
              }
              assert(false,
                  'GraphQLQueryX::buildQuery => Not supported field type "${x.runtimeType}"');
              return null;
            })
            .where((x) => x != null)
            .join(', ');
  }

  String build() {
    // build fields
    final retFields = _parseFields(fields);

    // build params
    final retParams = _parseParamMap(params);

    final retQueries = <String>[];
    final query = retParams == null
        ? '$name { $retFields }'
        : retFields.isNotEmpty
            ? '$name ( $retParams ) { $retFields }'
            : '$name ( $retParams )';

    retQueries.add(query);

    for (var subgql in _gqls) {
      retQueries.add(subgql.build());
    }

    return retQueries.join(', ');
  }

  String? _parseParamMap(Map<String, dynamic>? params) {
    if (params == null || params.isEmpty) return null;
    final retParams = params.entries
        .where((entry) => entry.value != null)
        .map((entry) =>
            '${paramCase(entry.key)}: ${_parseParamValue(entry.value)}')
        .toList();
    return retParams.join(', ');
  }

  String paramCase(String param) =>
      '${param[0].toLowerCase()}${param.substring(1)}';

  String _parseParamValue(dynamic value) {
    if (value is Map<String, dynamic>) {
      return '{ ${_parseParamMap(value)!} }';
    } else if (value is List<Map<String, dynamic>>) {
      return '[ ${value.map(_parseParamValue).join(',')} ]';
      //return value.map(_parseParamValue).join(',');
    } else if (value is int || value is num || value is bool) {
      return value.toString();
    } else if (value is Enum) {
      return _convertToGqlWords(value);
    } else {
      return '"$value"';
    }
  }

  // bool _isEnumItem(dynamic enumItem) {
  //   return enumItem is Enum;
  //   // final splitEnum = enumItem.toString().split('.');
  //   // return splitEnum.length > 1 &&
  //   //     splitEnum[0] == enumItem.runtimeType.toString();
  // }

  String _convertToGqlWords(dynamic enumItem) {
    final tmp = enumItem.toString().split('.')[1];
    return camelCaseToWords(tmp, separator: '_').toUpperCase();
  }

  Future<ApiResponse<T>> query<T>({
    Map<String, String>? headers,
    Map<String, dynamic>? extra,
    Duration? timeout,
    ApiResponseTransformer<T>? transformer,
  }) {
    return _gql<T>(
      'query',
      headers: headers,
      extra: extra,
      timeout: timeout,
      transformer: transformer,
    );
  }

  Future<ApiResponse<T>> mutation<T>({
    Map<String, String>? headers,
    Map<String, dynamic>? extra,
    Duration? timeout,
    ApiResponseTransformer<T>? transformer,
  }) {
    return _gql<T>(
      'mutation',
      headers: headers,
      extra: extra,
      timeout: timeout,
      transformer: transformer,
    );
  }

  Future<ApiResponse<T>> _gql<T>(
    String method, {
    Map<String, String>? headers,
    Map<String, dynamic>? extra,
    Duration? timeout,
    ApiResponseTransformer<T>? transformer,
  }) async {
    _headers ??= headers;
    _extra ??= extra;
    _timeout ??= timeout;
    _client ??= App.getIt<API>();
    _method = method;
    return _client!.gql<T>(method, this,
        options: Options(
          headers: _headers,
          sendTimeout: _timeout,
          extra: _extra,
        ),
        transformer: transformer);
  }
}

// @immutable
// class EnumSafeType {
//   final dynamic value;
//   const EnumSafeType(this.value);
//   String get gqlWords => EnumUtils.convertToGqlWords(value);
// }

// extension EnumSafeTypeHelper on Enum {
//   EnumSafeType get safeType => EnumSafeType(this);
// }
